import os

HOME_DIR = os.path.expanduser('~') + '/.FeelUOwn'

DATA_DIR = HOME_DIR + '/data'

USERS_INFO_FILE = DATA_DIR + '/nem_users_info.json'
NE_SOURCE = 'neteasemusic'
SONG_DIR = HOME_DIR + '/songs'

MUSIC_LIBRARY_PATH = os.path.expanduser('~') + '/Music'
